from pyspark.sql import SparkSession
from pyspark.sql import functions as F
from pyspark.sql.types import StructType, StructField, StringType, IntegerType, DoubleType, FloatType
from pyspark.sql.utils import AnalysisException
from pyspark.ml.regression import LinearRegression, LinearRegressionModel
from delta import configure_spark_with_delta_pip
from delta.tables import DeltaTable
from pyspark.ml.feature import VectorAssembler
import os
import re

# to set an entry null
def _set_null(item: str):
    return None if item == "null" else item

# move a file on the hdfs
def _move_hdfs_file(src: str, dest: str, spark: SparkSession): 
    Path = spark.sparkContext._gateway.jvm.org.apache.hadoop.fs.Path
    FileUtil = spark.sparkContext._gateway.jvm.org.apache.hadoop.fs.FileUtil
    conf = spark.sparkContext._jsc.hadoopConfiguration()

    src_path = Path(src)
    dest_path = Path(dest)
    FileUtil.copy(src_path.getFileSystem(conf), src_path, dest_path.getFileSystem(conf), dest_path, True, conf)

builder = SparkSession.builder.appName("LinearRegression")\
    .config("spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension")\
    .config("spark.sql.catalog.spark_catalog", "org.apache.spark.sql.delta.catalog.DeltaCatalog")\
    .config("spark.executor.instances", "3")\
    .config("spark.executor.cores", "4")\
    .config("spark.executor.memory", "4g")\
    .config("spark.driver.memory", "4g")
spark: SparkSession = configure_spark_with_delta_pip(builder).getOrCreate()

# paths for all diffrent directiories 
dataset_file_regex = "/dataset_rearranged/*.op"
namenode_dir = "hdfs://namenode:9000"
structured_dir = os.path.join(namenode_dir, "structured")
quarantine_dir = os.path.join(namenode_dir, "quarantine")
delta_dir = os.path.join(namenode_dir, "delta")
year_medians_dir = os.path.join(delta_dir, "year_medians")
lr_dir = os.path.join(delta_dir, "linear_regression")
structured_files = os.path.join(structured_dir, "*", "part-*")
delimiter = r'\s+'

# splitting lines at whitespaces, creating columns
def structure_line(line: str):
    splitted_line = re.split(delimiter, line)

    # Extract necessary columns and handle null values
    values = [
        _set_null(splitted_line[0]), _set_null(splitted_line[1]),
        _set_null(splitted_line[2]), _set_null(splitted_line[3]),
        _set_null(splitted_line[5]), _set_null(splitted_line[7]),
        _set_null(splitted_line[9]), _set_null(splitted_line[11]),
        _set_null(splitted_line[13]), _set_null(splitted_line[15]),
        _set_null(splitted_line[16]), _set_null(splitted_line[17]),
        _set_null(splitted_line[18]), _set_null(splitted_line[19]),
        _set_null(splitted_line[20]), _set_null(splitted_line[21])
    ]
    return ",".join(str(v) for v in values)

#read and apply structure_line
def preprocess_file(file: str):
    rdd = spark.sparkContext.textFile(file, minPartitions=81)
    result_rdd = rdd.map(structure_line) # !Serialization optimization! 
    output_path = os.path.join(structured_dir, os.path.basename(file) + ".csv")
    result_rdd.saveAsTextFile(output_path)
    return output_path

# apply schema, check schema, move invalid data, reducing to relevant columns, calculate median
def process_files(folder: str):
    files_path = os.path.join(folder, "*")
    
    schema = StructType([
        StructField("stn", IntegerType(), nullable=False),
        StructField("wban", IntegerType(), nullable=False),
        StructField("yearmoda", StringType(), nullable=False),
        StructField("temp", DoubleType(), nullable=False),
        StructField("dewp", DoubleType(), nullable=False),
        StructField("slp", DoubleType(), nullable=False),
        StructField("stp", DoubleType(), nullable=False),
        StructField("visib", DoubleType(), nullable=False),
        StructField("wdsp", DoubleType(), nullable=False),
        StructField("mxspd", DoubleType(), nullable=False),
        StructField("gust", DoubleType(), nullable=False),
        StructField("max", StringType(), nullable=False),
        StructField("min", StringType(), nullable=False),
        StructField("prcp", StringType(), nullable=False),
        StructField("sndp", DoubleType(), nullable=False),
        StructField("frshtt", StringType(), nullable=False)
    ])

    structured_data = spark.read.csv(files_path, header=False, schema=schema)
    structured_data = structured_data.withColumn("filename", F.input_file_name())

    # Check if data complies with the schema
    invalid_data = structured_data.filter(
        F.col("stn").isNull() | F.col("wban").isNull() | F.col("yearmoda").isNull() | F.col("temp").isNull() | F.col("dewp").isNull()
        | F.col("slp").isNull()  | F.col("stp").isNull() | F.col("visib").isNull() | F.col("wdsp").isNull() | F.col("mxspd").isNull() | F.col("gust").isNull()
        | F.col("max").isNull()| F.col("min").isNull() | F.col("prcp").isNull() | F.col("sndp").isNull()| F.col("frshtt").isNull()
    )

    invalid_paths = invalid_data.select("filename").dropDuplicates(["filename"])
    for file in invalid_paths.collect(): 
        path = file["filename"]
        _move_hdfs_file(path, quarantine_dir, spark)

    try: # if no files match structured_files, an AnalysisException is thrown
        structured_data = spark.read.csv(files_path, header=False, schema=schema) # must be newly read, since the removed data sources are still in the old structured_data and cannot be removed -> error when accessing data
    except AnalysisException as e:
        structured_data = spark.createDataFrame([], schema=schema)

    structured_data = structured_data.withColumn("date", F.to_date(F.col("yearmoda"), "yyyyMMdd")).drop("yearmoda")

    medians_pre_df = structured_data.select("stn", "temp", F.year("date").alias("year"))
    # !spill optimization! because of mmissing repartition
    medians_per_station_per_year = medians_pre_df.groupBy("stn", "year").agg(F.median("temp").alias("median_temp"))
    medians_per_year = medians_per_station_per_year.groupBy("year").agg(F.median("median_temp").alias("median"))
    medians_per_year = medians_per_year.select("year", "median")
    return medians_per_year

preprocessed_path = preprocess_file(dataset_file_regex)
medians_per_year = process_files(preprocessed_path)

# create delta table 
DeltaTable.createIfNotExists(spark).addColumns(medians_per_year.schema).location(year_medians_dir).execute()
yearly_median_table = DeltaTable.forPath(spark, year_medians_dir)
yearly_median_table.alias("oldData").merge(medians_per_year.alias("newData"), "oldData.year = newData.year")\
    .whenMatchedUpdate(set = {"year": "newData.year", "median": "newData.median"})\
    .whenNotMatchedInsert(values= {"year": "newData.year", "median": "newData.median"}).execute()

yearly_median_df = yearly_median_table.toDF()
yearly_median_df = yearly_median_df.sort("year")

# Execute linear regression
assembler = VectorAssembler(inputCols=["year"], outputCol="year_vec")
lr_data = assembler.transform(medians_per_year)
lr_data = lr_data.select("year_vec", "median")
lr = LinearRegression(featuresCol="year_vec", labelCol="median")
lr_model: LinearRegressionModel = lr.fit(lr_data)

# new schema for linear regression
lr_schema = StructType([
    StructField("coefficient", FloatType(), nullable=True),
    StructField("intercept", FloatType(), nullable=True)
])

# saving results of linear regression
lr_df = spark.createDataFrame([(lr_model.coefficients[0].item(), lr_model.intercept)], schema=lr_schema) # since lr only has one coefficient, lr_model.coefficients[0] is fine
lr_df.write.format("delta").mode("overwrite").save(lr_dir)
